
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

import itertools
import logging

from typing		import List, Optional

from shamir_mnemonic	import combine_mnemonics
from mnemonic		import Mnemonic

from ..util		import ordinal

from .entropy		import (  # noqa F401
    shannon_entropy, signal_entropy, analyze_entropy, scan_entropy, display_entropy
)

log				= logging.getLogger( __package__ )


def recover(
    mnemonics: List[str],
    passphrase: bytes		= b"",
    using_bip39: bool		= False,
) -> bytes:
    """Recover a master secret Seed Entropy from the supplied SLIP-39 mnemonics.  We cannot know what
    subset of these mnemonics is required and/or valid, so we need to iterate over all subset
    combinations on failure.

    We'll try to find one of the smallest subsets that satisfies the SLIP-39 recovery.  The
    resultant secret Entropy is returned as the Seed, with (not widely used) SLIP-39 decryption with
    the given passphrase.

    WARNING: SLIP-39 passphrase encryption is not Trezor "Model T" compatible, and is not widely
    used; if you want to hide a wallet, use the Trezor "Hidden wallet" feature instead, where the
    passphrase for each hidden wallet is entered on the device (leave this passphrase blank).

    Optionally, if using_bip39 then generate the Seed from the recovered Seed Entropy, using BIP-39
    Seed generation.  This is the ideal way to back up an existing BIP-39 Mnemonic, I believe.  It
    results in a 20- or 33-word SLIP-39 Mnemonics, retains the BIP-39 passphrase(s) for securing the
    resultant Cryptocurrency wallet(s), and is compatible with all BIP-39 hardware wallets.  Once
    you are comfortable that you can always recover your BIP-39 Mnemonic from your SLIP-39 Mnemomnic
    Cards, you are free to destroy your original insecure and unreliable BIP-39 Mnemonic backup(s).

    """
    secret			= None
    try:
        combo			= range( len( mnemonics ))
        secret			= combine_mnemonics(
            mnemonics,
            passphrase	= b"" if using_bip39 else passphrase
        )
    except Exception as exc:
        # Try a subset of the supplied mnemonics, to silently reject any invalid mnemonic phrases supplied
        for length in range( len( mnemonics )):
            for combo in itertools.combinations( range( len( mnemonics )), length ):
                trial		= list( mnemonics[i] for i in combo )
                try:
                    secret	= combine_mnemonics(
                        trial,
                        passphrase	= b"" if using_bip39 else passphrase
                    )
                    break
                except Exception:
                    pass
            if secret:
                break
        if not secret:
            # No recovery; raise the Exception produced by original attempt w/ all mnemonics
            raise exc
    log.info(
        f"Recovered {len(secret)*8}-bit SLIP-39 Seed Entropy with {len(combo)}"
        f" ({'all' if len(combo) == len(mnemonics) else ', '.join( ordinal(i+1) for i in combo)})"
        f" of {len(mnemonics)} supplied mnemonics" + (
            f"; Seed decoded from SLIP-39 (w/ no passphrase) and generated using BIP-39 Mnemonic representation w/ {'a' if passphrase else 'no'} passphrase"
            if using_bip39 else
            f"; Seed decoded from SLIP-39 Mnemonics w/ {'a' if passphrase else 'no'} passphrase"
        )
    )
    if using_bip39:
        secret			= recover_bip39(
            mnemonic	= produce_bip39( entropy=secret ),
            passphrase	= passphrase,
        )
    return secret


def recover_bip39(
    mnemonic: str,
    passphrase: bytes		= b"",
    as_entropy			= False,  # Recover original 128- or 256-bit Entropy (not 512-bit Seed)
) -> bytes:
    """Recover a secret 512-bit BIP-39 generated seed (or just the original 128- or 256-bit entropy)
    from a single BIP-39 mnemonic phrase, detecting the language.

    """
    assert not ( bool( passphrase ) and bool( as_entropy )), \
        "When recovering original BIP-39 entropy, no passphrase may be specified"
    # Polish up the supplied mnemonic, by eliminating extra spaces; Mnemonic is fragile...
    mnemonic			= ' '.join( w.lower() for w in mnemonic.split( ' ' ) if w )
    # Unfortunately, Mnemonic.detect_language was unreliable; only checked the first word
    # and english/french has ambiguous words. TODO: check if python-mnemonic has been fixed.
    last			= ValueError( "Empty mnemonic" )
    for word in mnemonic.split( ' ' ):
        try:
            language		= Mnemonic.detect_language( word )
            m			= Mnemonic( language )
            assert m.check( mnemonic ), \
                f"Invalid {language} mnemonic: {mnemonic!r}"
            if as_entropy:
                secret		= m.to_entropy( mnemonic )
            else:
                secret		= Mnemonic.to_seed( mnemonic, passphrase )
            log.info( f"Recovered {len(secret)*8}-bit BIP-39 secret from {language} mnemonic" )
            return bytes( secret )  # bytearray --> bytes
        except Exception as exc:
            last		= exc
    raise last


def produce_bip39(
    entropy: Optional[bytes],
    strength: Optional[int]	= None,
    language: str		= "english",
) -> str:
    """Produce a BIP-38 Mnemonic from the provided entropy (or generated, default 128 bits)."""
    mnemo			= Mnemonic( language )
    if entropy:
        return mnemo.to_mnemonic( entropy )
    return mnemo.generate( strength or 128 )




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

import colorsys
import fractions
import getpass
import logging
import math
import sys

from functools		import wraps

# util.timer
#
# Select platform appropriate timer function
#
if sys.platform == 'win32' and sys.version_info[0:2] < (3,8):
    # On Windows (before Python 3.8), the best timer is time.clock
    from time		import clock	as timer
else:
    # On most other platforms the best timer is time.time
    from time		import time	as timer


#
# @util.timing
#
def timing( fun, instrument=False, level=None ):
    """A timing decorator which optionally instruments the function return value to return
    (duration,value), or logs the call duration (at logging.INFO by default, if not instrumenting).

    """
    @wraps( fun )
    def wrap( *args, **kwds ):
        beg			= timer()
        result			= fun( *args, **kwds )
        end			= timer()
        dur			= end - beg
        if level or not instrument:
            logging.log( level or logging.INFO, f"{fun!r} took {dur:.3f}s" )
        if instrument:
            return (dur,result)
        return result
    return wrap


log_cfg				= {
    "level":	logging.WARNING,
    "datefmt":	'%Y-%m-%d %H:%M:%S',
    #"format":	'%(asctime)s.%(msecs).03d %(threadName)10.10s %(name)-16.16s %(levelname)-8.8s %(funcName)-10.10s %(message)s',
    "format":	'%(asctime)s %(name)-16.16s %(message)s',
}

log_levelmap 			= {
    -2: logging.FATAL,
    -1: logging.ERROR,
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def log_level( adjust ):
    """Return a logging level corresponding to the +'ve/-'ve adjustment"""
    return log_levelmap[
        max(
            min(
                adjust,
                max( log_levelmap.keys() )
            ),
            min( log_levelmap.keys() )
        )
    ]


def ordinal( num ):
    ordinal_dict		= {1: "st", 2: "nd", 3: "rd"}
    q, mod			= divmod( num, 10 )
    suffix			= q % 10 != 1 and ordinal_dict.get(mod) or "th"
    return f"{num}{suffix}"


def commas( seq, final_and=None ):
    """Replace any numeric sequences eg. 1, 2, 3, 5, 7 w/ 1-3, 5 and 7.  Caller should
    usually sort numeric values before calling."""
    def int_seq( seq ):
        for i,iv in enumerate( seq[:-1] ):
            if type(iv) in (int,float):
                for j,jv in enumerate( seq[i:] ):
                    if type(jv) not in (int,float) or jv != iv + j:
                        j      -= 1
                        break
                if j > 1:
                    return (i,i+j)
        return None
    seq				= list( seq )
    while rng := int_seq( seq ):
        #print( f"int_seq: series {rng!r} ({seq[rng[0]]!r} - {seq[rng[1]]!r}) found in {seq!r}"  )
        beg			= seq[:rng[0]]
        nxt			= rng[1] + 1
        end			= seq[nxt:] if nxt < len( seq ) else []
        seq			= beg + [f"{seq[rng[0]]}-{seq[rng[1]]}"] + end
    if final_and and len(seq) > 1:
        seq			= seq[:-2] + [f"{seq[-2]} and {seq[-1]}"]
    return ', '.join( map( str, seq ))


def round_onto( value, keys, keep_sign=True ):
    """Find the numeric key closest to value, maintaining sign if possible.  None and other
    non-numeric values are supported if they are present in keys.

    """
    if value in keys:
        return value
    keys			= sorted( k for k in keys if type(k) in (float,int) )
    near, near_i		= min(
        (abs( k - value), i)
        for i, k in enumerate( keys )
    )
    if keep_sign and (( value < 0 ) != ( keys[near_i] < 0 )):
        # The sign differs; if value -'ve but closest was +'ve, and there is a lower key available
        if value < 0:
            if near_i > 0:
                near_i	       -= 1
        else:
            if near_i + 1 < len( keys ):
                near_i	       += 1
    return keys[near_i]


entropy_signal_strengths	= {
    3: "very bad",
    2: "bad",
    1: "poor",
    0: "weak",
    -1: "ok",
    -2: "strong",
    None: "excellent",
}


def entropy_rating_dB( dB ):
    return entropy_signal_strengths[round_onto( dB, entropy_signal_strengths.keys(), keep_sign=True )]


def rate_dB( dB, what=None ):
    rating			= entropy_rating_dB( dB )
    result			= ''
    if what:
        result		       += f"{what}: "
    result		       += f"{dB:.1f}dB" if type( dB ) in (int,float) else f"{dB}"
    result		       += f" ({rating})"
    return result


def input_secure( prompt, secret=True, file=None ):
    """When getting secure (optionally secret) input from standard input, we don't want to use getpass, which
    attempts to read from /dev/tty.

    """
    if ( file or sys.stdin ).isatty():
        # From TTY; provide prompts, and do not echo secret input
        if secret:
            return getpass.getpass( prompt, stream=file )
        elif file:
            # Coming from some file; no prompt, read a line from the file source
            return file.readline()
        else:
            return input( prompt )
    else:
        # Not a TTY; don't litter pipeline output with prompts
        if file:
            return file.readline()
        return input()


def chunker( sequence, size ):
    while sequence:
        yield sequence[:size]
        sequence		= sequence[size:]


def hex_to_rgb( value, real=False, precision=4 ):
    """
    Convert hex color to ints, or reals rounded to a certain precision.

    >>> hex_to_rgb( "#ffffff" )
    (255, 255, 255)
    >>> hex_to_rgb( "#ffffffffffff" )
    (65535, 65535, 65535)
    >>> hex_to_rgb( "#ffffffffffff", real=True )
    (1.0, 1.0, 1.0)
    >>> hex_to_rgb( "#ffff80002000", real=True )
    (1.0, 0.5, 0.125)
    """
    nibble			= value.lstrip('#')
    nibs			= len( nibble )
    ints			= ( int( nibble[i:i + nibs // 3], 16 ) for i in range( 0, nibs, nibs // 3 ))
    return tuple( ( round( i / ( 16 ** ( nibs // 3 ) - 1 ), precision ) for i in ints ) if real else ints )


def rgb_to_hex( *rgb, bits=None, nibs=None ):
    """Convert int/real RGB values to hex.  For ints, assumes 8-bit values in range [0-255); for
    [0,65535), use bits=16.  For real-valued rgb in range (0,1), use bits=1.

    By default, will try to auto-select between bits=0/8/16, by examining rgb values; if all are <1,
    or <=1 and are float, then bits=0; otherwise if any are > 255, then bits=16.

    >>> rgb_to_hex( 255, 255, 255 )
    '#ffffff'
    >>> rgb_to_hex( 65535, 65535, 65535 )
    '#ffffffffffff'

    >>> rgb_to_hex( .5, .5, .5 )		# deduces (0,1)
    '#808080'

    >>> rgb_to_hex( .5, .75, .9, nibs=3  )	# same, but force 3 nibbles
    '#800bffe66'

    >>> rgb_to_hex( 11, 111, 1111 )		# deduces (0,65535]
    '#000b006f0457'

    >>> rgb_to_hex( 288, 255, 254, bits=8 )	# force 8-bit colors, w/ truncation
    '#fffffe'
    """
    assert len( rgb ) == 3, \
        "Requires RGB values"
    if bits is None:
        if all( c < 1 for c in rgb ) or all( c <= 1 and type( c ) is float for c in rgb ):
            bits		= 1
        elif any( c > 255 for c in rgb ):
            bits		= 16
        else:
            bits		= 8

    if nibs is None:
        nibs			= 2 if bits == 1 else bits // 4 + ( 1 if ( bits % 4 ) else 0 )

    end				= 2 ** bits - 1
    reals			= tuple( c / end for c in rgb )
    one				= 16 ** nibs - 1
    ints			= ( max( 0, min( one, int( round( r * one )))) for r in reals )

    return '#' + ''.join( "{i:0{nibs}x}".format( nibs=nibs, i=i ) for i in ints )


def hue_shift( color, shift=1/3 ):
    """Take a hex or (r,g,b) value, and shift its Hue "right" in the HSV spectrum by the given shift
    percentage (default: .1, or 10%).  Requires either '#aabbcc' hex colors, or real-values RGB in
    the range (0,1).

    >>> hue_shift( '#ff0000', shift=0 )
    '#ff0000'

    >>> hue_shift( '#ff0000', shift=1/3 )
    '#00ff00'

    >>> hue_shift( '#ff0000', shift=-1/3 )
    '#0000ff'

    >>> hue_shift( '#ff0000', shift=2/3 )
    '#0000ff'

    >>> hue_shift( '#ff0000', shift=1/2 )
    '#00ffff'

    >>> hue_shift( '#ff0000', shift=1/4 )
    '#80ff00'

    >>> hue_shift( '#ff0000', shift=1/10 )
    '#ff9900'

    >>> hue_shift( '#800000', shift=1/10 )
    '#804d00'

    >>> hue_shift( '#48c', shift=1/3 )
    '#cc4488'

    >>> hue_shift( (4/15, 8/15, 12/15), shift=-2/3 )  # Same RGB, but Hue shifted the other way 'round
    '#cc4488'
    """
    if isinstance( color, str ):
        rgb			= hex_to_rgb( color, real=1 )
    else:
        rgb			= tuple( color )
    assert len( rgb ) == 3 and all( 0.0 <= c <= 1.0 for c in rgb ), \
        f"Invalid RGB values for hue shift: {rgb!r}"
    h,s,v			= colorsys.rgb_to_hsv( *rgb )
    #log.warning( f"Color {color!r} == RGB {rgb!r} == HSV{(h,s,v)!r}" )
    h				= ( h + shift ) % 1.0  # if shift is -'ve, % (modulo) correctly shifts to +'ve
    rgb				= colorsys.hsv_to_rgb( h, s, v )
    return rgb_to_hex( *rgb )


def exponential_moving_average( current, sample, weight ):
    """exponential_moving_average -- rolling average without any data history

    Computes an exponential moving average:

        ( 1 - weight ) * current + weight * sample

    where the incoming sample has the given weight, and current samples have exponentially less
    influence on the current value.  Ignores a current value of None.

    """
    return sample if current is None else current + weight * ( sample - current )


def avg( seq ):
    vals			= list( seq )
    if vals:
        return sum( vals ) / len( vals )
    return math.nan


def rms( seq ):
    """Computes RMS for real/complex sequence of values"""
    if seq:
        return math.sqrt( avg( abs( s ) ** 2 for s in seq ))
    return 0


def is_power_of_2( n: int ) -> bool:
    return not ( n & ( n - 1 ))


class mixed_fraction( fractions.Fraction ):
    """A Fraction that represents whole multiples of itself as eg. 1-1/2"""
    def __str__( self ):
        whole, rest		= divmod( self.numerator, self.denominator )
        if whole and rest:
            return f"{whole}+{fractions.Fraction( rest, self.denominator)}"
        return super().__str__()
